package collections

import (
	"testing"

	"git.perx.ru/perxis/perxis-go/pkg/optional"
	"git.perx.ru/perxis/perxis-go/pkg/schema"
	"git.perx.ru/perxis/perxis-go/pkg/schema/field"
	"github.com/stretchr/testify/require"
)

func TestView_Equal(t *testing.T) {
	testCases := []struct {
		name  string
		view1 *View
		view2 *View
		want  bool
	}{
		{
			name:  "Both views are nil",
			view1: nil,
			view2: nil,
			want:  true,
		},
		{
			name:  "View1 is nil",
			view1: nil,
			view2: &View{},
			want:  false,
		},
		{
			name:  "View2 is nil",
			view1: &View{},
			view2: nil,
			want:  false,
		},
		{
			name: "Different views",
			view1: &View{
				SpaceID:      "spaceID1",
				EnvID:        "envID1",
				CollectionID: "collectionID1",
				Filter:       "filter1",
			},
			view2: &View{
				SpaceID:      "spaceID2",
				EnvID:        "envID2",
				CollectionID: "collectionID2",
				Filter:       "filter2",
			},
			want: false,
		},
		{
			name: "Equal views",
			view1: &View{
				SpaceID:      "spaceID",
				EnvID:        "envID",
				CollectionID: "collectionID",
				Filter:       "filter",
			},
			view2: &View{
				SpaceID:      "spaceID",
				EnvID:        "envID",
				CollectionID: "collectionID",
				Filter:       "filter",
			},
			want: true,
		},
		{
			name:  "Empty view1 and non-empty view2",
			view1: &View{},
			view2: &View{
				SpaceID:      "spaceID",
				EnvID:        "envID",
				CollectionID: "collectionID",
				Filter:       "filter",
			},
			want: false,
		},
		{
			name: "Empty view2 and non-empty view1",
			view1: &View{
				SpaceID:      "spaceID",
				EnvID:        "envID",
				CollectionID: "collectionID",
				Filter:       "filter",
			},
			view2: &View{},
			want:  false,
		},
		{
			name:  "Empty view1 and view2",
			view1: &View{},
			view2: &View{},
			want:  true,
		},
	}

	for _, tc := range testCases {
		t.Run(tc.name, func(t *testing.T) {
			output := tc.view1.Equal(tc.view2)
			require.Equal(t, tc.want, output)
		})
	}
}

func TestFromSchemaMetadata(t *testing.T) {
	testCases := []struct {
		name    string
		schemas []*schema.Schema
		want    []*Collection
	}{
		{
			name:    "Nil",
			schemas: nil,
			want:    []*Collection{},
		},
		{
			name:    "Empty",
			schemas: []*schema.Schema{},
			want:    []*Collection{},
		},
		{
			name:    "Without metadata",
			schemas: []*schema.Schema{schema.New("a", field.String())},
			want:    []*Collection{{Schema: schema.New("a", field.String())}},
		},
		{
			name: "With metadata",
			schemas: []*schema.Schema{schema.New("a", field.String()).WithMetadata(
				"collection_id", "collID",
				"collection_name", "collName",
				"collection_single", "true",
				"collection_system", "true",
				"collection_nodata", "true",
				"collection_hidden", "true",
				"collection_view_space", "viewSpaceID",
				"collection_view_env", "viewEnvID",
				"collection_view_id", "viewCollID",
				"collection_view_filter", "viewFilter",
			)},
			want: []*Collection{{
				ID:     "collID",
				Name:   "collName",
				Single: optional.True,
				System: optional.True,
				NoData: optional.True,
				Hidden: true,
				Schema: schema.New("a", field.String()).WithMetadata("collection_id", "collID", "collection_name", "collName", "collection_single", "true", "collection_system", "true", "collection_nodata", "true", "collection_hidden", "true", "collection_view_space", "viewSpaceID", "collection_view_env", "viewEnvID", "collection_view_id", "viewCollID", "collection_view_filter", "viewFilter"),
				View: &View{
					SpaceID:      "viewSpaceID",
					EnvID:        "viewEnvID",
					CollectionID: "viewCollID",
					Filter:       "viewFilter",
				},
			}},
		},
		{
			name: "Multiple",
			schemas: []*schema.Schema{
				schema.New("a", field.String()).WithMetadata("collection_id", "collID"),
				schema.New("b", field.String()).WithMetadata("collection_name", "collName"),
			},
			want: []*Collection{
				{
					ID:     "collID",
					Name:   "",
					Schema: schema.New("a", field.String()).WithMetadata("collection_id", "collID"),
				},
				{

					ID:     "",
					Name:   "collName",
					Schema: schema.New("b", field.String()).WithMetadata("collection_name", "collName"),
				},
			},
		},
	}

	for _, tt := range testCases {
		t.Run(tt.name, func(t *testing.T) {
			result := FromSchemaMetadata(tt.schemas...)
			require.Equal(t, tt.want, result)
		})
	}
}
